#include <IOKit/IOLib.h>
#include <IOKit/pci/IOPCIBridge.h>
#include "pcireg.h"
#include "agpreg.h"

#define A_SIZE_16(x)    ((struct aper_size_info_16 *) x)
#define AGP_PAGE_SHIFT          12
#define kDevice_IDProperty			"device-id"
#define kVendor_IDProperty			"vendor-id"

/* Definitions of PCI2PCI Config Registers */
enum {
    kPCI2PCIPrimaryBus		= 0x18,		// 8 bit
    kPCI2PCISecondaryBus	= 0x19,		// 8 bit
    kPCI2PCISubordinateBus	= 0x1a,		// 8 bit
    kPCI2PCIIORange		= 0x1c,		
    kPCI2PCIMemoryRange		= 0x20,		
    kPCI2PCIPrefetchMemoryRange	= 0x24,		
    kPCI2PCIUpperIORange	= 0x30			
};

enum {
    kIOAGP3Mode			= 0x00000008,
    kIOAGP4xDataRateMode3	= 0x00000001,
    kIOAGP8xDataRateMode3	= 0x00000002,
};

/* Our 5 currently supported vendors */
enum{
		AMD64,
		ATI,
		INTEL,
		VIA,
		SiS
};

struct aper_size_info_8 {
          int size;
          int num_entries;
          int page_order;
          UInt8 size_value;
};

struct aper_size_info_16 {
          int size;
          int num_entries;
          int page_order;
          UInt16 size_value;
};
 
static inline void agp_flush_cache()
{
	__asm __volatile("wbinvd");
}
//static volatile UInt32*	gartArray;
//static UInt32			gartPhys;
//static OSData *			gartHandle;

#define pciread32(space,offset)				bridge->configRead32(space,offset)
#define pciread16(space,offset)				bridge->configRead16(space,offset)
#define pciread8(space,offset)				bridge->configRead8(space,offset)
#define pciwrite32(space,offset,value)		bridge->configWrite32(space,offset,value)
#define pciwrite16(space,offset,value)		bridge->configWrite16(space,offset,value)
#define pciwrite8(space,offset,value)		bridge->configWrite8(space,offset,value)

class IOAGPBridge : public OSObject
{	
	OSDeclareDefaultStructors(IOAGPBridge)
	
	public:
		UInt8						targetAGPRegisters;
		IOPCIAddressSpace			hostSpace;
		IOPCIAddressSpace			AGPBSpace;
		UInt32				addrMask;
		UInt32				ITA_COH;
		UInt32				gartWidth;
		bool				isAMD;
		int major_versionM, minor_versionM;
		UInt32				AGPFlags;
		UInt8 first, last, cacheline, mingnt, lattimer, maxlat, irq;
		UInt16 stat, secstat, cmd;
		UInt16 base, limit, pbase, plimit;
		UInt8 iobase, iolimit, devsel, fastback;
		
		// GART
	//	volatile UInt32*			gartArray;
	//	UInt32			gartPhys;
	//	UInt32			entries;
		UInt32			systemLength;
		UInt32			systemBase;
		UInt32			initial_aperture;
		UInt32			apbase;
		IOPCIDevice		*bridge;
	//	OSData * gartHandle; 
	//	volatile UInt32* agpMemory;
	//	UInt32			agpPhys;

		// Used by Intel driver
		void			*current_size;
		UInt16			apsizeValue;
		IORangeAllocator *agpRange;
		IOByteCount		gartLength;
		// Used by VIA driver
		UInt8			*regs;
		int				agpver;
		bool			readyForPM;
		int						n_mctrl;
		int						mctrl[256];
		//int						via_agp;
	
		 // Used by ATI driver
		 UInt8 apsize_reg;
		 UInt8 agpmode_reg;
		 IOMemoryMap	*mmio;
		volatile UInt8 *mmio_base;
		unsigned long   mmio_base_phys;
		//volatile UInt8 *ATIregisters; //__iomem 
		IOMemoryMap	*RADio;
		volatile UInt8 *RADio_base;
		
		UInt32	flusht;
		UInt32 tstatus;
		UInt32 mstatus;

	public:
	
		virtual const char* Match(UInt16 dev);
		void printBridgeConfig(IOPCIDevice * AGPBridge);
		IODeviceMemory *getVRAMMemory(IOAGPDevice * master);
		virtual IOReturn Configure(IOAGPDevice *master, IOPhysicalAddress * address, IOPhysicalLength * length);
		virtual IOReturn InstallGATT(IOPhysicalAddress gattPhys);
		virtual void Flush0TLB();
		virtual void Flush1TLB();
		virtual void CheckTLB();
//		virtual IOReturn BindMemory(IOMemoryDescriptor * mem, IOByteCount agpOffset);
		virtual int GetAGPVersion(IOAGPDevice * master);
		IOReturn Enable(IOAGPDevice * master, UInt32 mode);
		virtual void Disable( IOAGPDevice * master );
		
	private:
		IOReturn EnableMode3(IOAGPDevice * agpMaster, UInt32 mode);
		IOReturn EnableMode2(IOAGPDevice * agpMaster, UInt32 mode);
				
};

class AGPAMD64 : public IOAGPBridge
{
	public:
		AGPAMD64(IOService *provider);
		const char* Match(UInt16 dev);
		IOReturn Configure(IOAGPDevice *master, 
						IOPhysicalAddress * address, IOPhysicalLength * length);
		IOReturn InstallGATT(IOPhysicalAddress gattPhys);
		void Flush0TLB();
		void Flush1TLB();
		void CheckTLB();
		//int BindMemory(IOMemoryDescriptor * mem, IOByteCount agpOffset);
		void Disable( IOAGPDevice * master );
	private:
		void agp_amd64_apbase_fixup();
		int agp_amd64_via_match(void);
		void agp_amd64_via_init();
		int agp_amd64_via_set_aperture(UInt32 aperture);
		int agp_amd64_nvidia_match(UInt16 devid);
		UInt32 agp_amd64_get_aperture(IOPCIAddressSpace dev);
		void agp_amd64_uli_init();
		int agp_amd64_uli_set_aperture(UInt32 aperture);
		void agp_amd64_nvidia_init();
		int agp_amd64_nvidia_set_aperture(UInt32 aperture);
		void agp_amd64_ATI_init();
		int agp_amd64_ATI_set_aperture(UInt32 aperture);

};

class AGPVIA : public IOAGPBridge
{
	public:
		AGPVIA(IOService *provider);
		const char* Match(UInt16 dev);
//		IOReturn BindMemory(IOMemoryDescriptor * mem, IOByteCount agpOffset);
		IOReturn Configure(IOAGPDevice *master, 
						IOPhysicalAddress * address, IOPhysicalLength * length);
		IOReturn InstallGATT(IOPhysicalAddress gattPhys);
		void Flush0TLB();
		void Flush1TLB();
		void CheckTLB();
		void Disable( IOAGPDevice * master );
	private:
		UInt32 agp_via_get_aperture();		
};

class AGPINTEL : public IOAGPBridge
{
	public:
		AGPINTEL(IOService *provider);
		const char* Match(UInt16 dev);
//		IOReturn BindMemory(IOMemoryDescriptor * mem, IOByteCount agpOffset);
		IOReturn Configure(IOAGPDevice *master, 
						IOPhysicalAddress * address, IOPhysicalLength * length);
		IOReturn InstallGATT(IOPhysicalAddress gattPhys);
		void Flush0TLB();
		void Flush1TLB();
		void CheckTLB();
		void Disable( IOAGPDevice * master );
	private:
		UInt32 GetBestAPBASE(IOAGPDevice *_master, UInt32 currentAperBase);
};

class AGPATI : public IOAGPBridge
{
	public:
		AGPATI(IOService *provider);
		const char* Match(UInt16 dev);
//		IOReturn BindMemory(IOMemoryDescriptor * mem, IOByteCount agpOffset);
		IOReturn Configure(IOAGPDevice *master, 
						IOPhysicalAddress * address, IOPhysicalLength * length);
		IOReturn InstallGATT(IOPhysicalAddress gattPhys);
		void Flush0TLB();
		void Flush1TLB();
		void CheckTLB();
		void Disable( IOAGPDevice * master );
	private:
		UInt32 GetBestAPBASE(IOAGPDevice *_master, UInt32 currentAperBase);
};
