/*
 * Copyright (c) 1998-2000 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
 // 

#include <IOKit/IOLib.h>
#include <IOKit/platform/ApplePlatformExpert.h>
#include <IOKit/pci/IOPCIBridge.h>
#include <IOKit/IOMapper.h>

#include "IOAGPBridge.h"


#define  PCI_BASE_ADDRESS_MEM_MASK      (~0x0fUL)
#define ATTBASE_MASK 0x00000FFF
#define PCI_STATE		0x06		bit 4 = 1 cap_ptr presents

//Slice - experimental registers
#define ATI_MEM_RANGE_1		0x24
#define ATI_IO_RANGE_1		0x20
#define ATI_RAM_SIZE		0x90
#define ATI_AGP_MODE_PTR	0x9c

//From agp3.pdf Registers in PCI space
#define APBASE			0x10
#define CAPABILITY_PTR	0x34

// Then shift in AGP space
#define AGP_STAT			0x04
#define AGP_CMD				0x08
#define AGP_CTRL			0x10
#define APSIZE				0x14 
#define ATTBASE				0x18

// ATI specific
//----------------------------------------------------------------------------
 #define ATI_GART_MMBASE_ADDR    0x14
 #define ATI_RS100_APSIZE        0xac
 #define ATI_RS100_IG_AGPMODE    0xb0
 #define ATI_RS300_APSIZE        0xf8
 #define ATI_RS300_IG_AGPMODE    0xfc
 // mmio space
 #define ATI_GART_FEATURE_ID             0x00
 #define ATI_GART_BASE                   0x04
 #define ATI_GART_CACHE_SZBASE           0x08
 #define ATI_GART_CACHE_CNTRL            0x0c
 #define ATI_GART_CACHE_ENTRY_CNTRL      0x10
 // Radeon space
 #define RADEON_BUS_CNTL					 0x30
 #define RADEON_AGP_BASE					0x170
 #define RADEON_AGP_CNTL					0x174

/********************************************************************/
/*						INTEL DRIVER FUNCTIONS						*/
/*						 by joblo10 and Tom H	
						 August 8, 2006
						 Slice: Intel -> ATI						*/
/*						   		September, 2007						*/	
/********************************************************************/

#define super IOAGPBridge

#define INREG(offset) OSReadLittleInt32((mmio_base), offset)
#define OUTREG(offset,val) OSWriteLittleInt32((mmio_base), offset, val)

#define INRAD(offset) OSReadLittleInt32((RADio_base), offset)
#define OUTRAD(offset,val) OSWriteLittleInt32((RADio_base), offset, val)

//----------------------------------------------------------------------------

AGPATI::AGPATI(IOService *provider)
{
	//Accessor for configRead and configWrite 
	bridge = (IOPCIDevice *)provider;
	AGPBSpace = bridge->space;
	//PCI Space of the host bridge
	hostSpace = AGPBSpace;
	hostSpace.s.deviceNum = 0;	
}

//----------------------------------------------------------------------------

const char* AGPATI::Match(UInt16 dev)
{
	// AGBP space?
		switch (dev) {
		case 0xcab0:
			return ("ATI RS100 AGP bridge");
		case 0xcab2:
			return ("ATI RS200 AGP bridge");
		case 0xcbb2:
		case 0x7010:
			return ("ATI RS200M AGP bridge");
		case 0xcab3:
			return ("ATI RS250 AGP bridge");
		case 0x5830:
			return ("ATI RS300_100 AGP bridge");
		case 0x5831:
			return ("ATI RS300_133 AGP bridge");
		case 0x5832:
			return ("ATI RS300_166 AGP bridge");
		case 0x5833:
			return ("ATI RS300_200 AGP bridge");
		case 0x5838:		//Slice
			return ("ATI RS300M AGP bridge");
		case 0x5a3f:		//Slice - this is AMD
			return ("ATI RS400 AGP bridge");
		case 0x7932:
			return ("ATI RS480 AGP bridge");
		default:
			return NULL;

	};

	 return NULL;
} 

//----------------------------------------------------------------------------

static struct 
aper_size_info_16 ati_generic_sizes[] =
{
         {256, 65536, 6, 0xf00},
         {128, 32768, 5, 0xf20},
         {64, 16384, 4, 0xf30},
         {32, 8192, 3, 0xf38},
         {16, 4096, 2, 0xf3c},
         {8, 2048, 1, 0xf3e},
         {4, 1024, 0, 0xf3f}
};

UInt32 AGPATI::GetBestAPBASE(IOAGPDevice *_master, UInt32 currentAperBase)
{
/*	//IODeviceMemory *mem = getVRAMMemory(_master);
	UInt32 VRAMBase, VRAMLength, gap, tom, temp;
	IOLog("AGPATI: Get BestAPBASE\n");
	//VRAMBase = mem->getPhysicalAddress();	// 0xe4000000
	//VRAMLength = mem->getLength();			// 0x02000000 - 32Mb - might be but I have 0x04000000
	tom = INRAD(0x15c);
	VRAMBase = (tom & 0xffff) << 16;
	VRAMLength = ( tom & 0xffff0000) - VRAMBase + 0x10000;
	//OUTRAD(0x148, tom);
	//OUTRAD(0x23c, VRAMBase);
	//OUTRAD(0x33c, VRAMBase);
	
	temp = VRAMBase;
	//temp = VRAMBase + VRAMLength; // + INRAD(0xf8);
	//First, verify if there's a GAP between VRAM and AGP aperture
	//temp = bridgeBase;
	IOLog("AGPATI: VRAM = [%08lx, %08lx]\n", VRAMBase, VRAMLength);
	if(VRAMBase == 0) VRAMBase = 0xe4000000;
		if(VRAMLength == 0) VRAMLength = 0x02000000;
	
	gap = currentAperBase - (VRAMBase + VRAMLength);
	
	if(0)  //currentAperBase > (VRAMBase + VRAMLength))
	{
		//gap exist and agp aperture is higher than VRAM, APBASE will be reduced
		temp = currentAperBase - gap;
		IOLog("AGP: Gap reduced by %08lx to zero\n", gap);
		//verify it's a valid value
		if(temp & 0x00FFFFFF)
			IOLog("AGP: New APBASE invalid\n");
		else
	} */
	IODeviceMemory *mem = getVRAMMemory(_master);
	UInt32 VRAMBase, VRAMLength, gap, temp;
	
	VRAMBase = mem->getPhysicalAddress();
	VRAMLength = mem->getLength() >> 1;
	//temp = currentAperBase;
	temp = VRAMBase + VRAMLength;
	//First, verify if there's a GAP between VRAM and AGP aperture
	gap = currentAperBase - temp; //(VRAMBase + VRAMLength);
	if(currentAperBase > temp) //(VRAMBase + VRAMLength))
	{
		//gap exist and agp aperture is higher than VRAM, APBASE will be reduced
		temp = currentAperBase - gap;
		IOLog("AGP: Gap exists, aperBase corrected\n");
		//verify it's a valid value
		if(temp & 0x00FFFFFF)
		{
			IOLog("AGP: New APBASE invalid, return default value\n");
			return currentAperBase;
		}
	}
	
		return temp;
}

//----------------------------------------------------------------------------
#define kIOServicePlane			"IOService"

IOReturn AGPATI::Configure(IOAGPDevice *master, IOPhysicalAddress * address, IOPhysicalLength * length)
{
    UInt32	VRAM=0, rVRAM; //, temp, temp2;
	UInt32 vendMast;
	UInt32	RADio_base_phys;
	aper_size_info_16 *values;
	IOReturn err = kIOReturnSuccess;
	IOLog("AGPATI configure init length %08lx\n", *length);
		// Find chipset type
	switch (pciread32(hostSpace, 0)) {
         case 0xcab01002: /* ATI RS100 AGP bridge */
         case 0xcab21002: /* ATI RS200 AGP bridge */
         case 0xcbb21002: /* ATI RS200M AGP bridge */
         case 0xcab31002: /* ATI RS250 AGP bridge */
                 IOLog("AGPATI: Chipset is < RS300.\n");
                 apsize_reg = ATI_RS100_APSIZE;
                 agpmode_reg = ATI_RS100_IG_AGPMODE;
                 break;
         case 0x58301002: /* ATI RS300_100 AGP bridge */
         case 0x58311002: /* ATI RS300_133 AGP bridge */
         case 0x58321002: /* ATI RS300_166 AGP bridge */
         case 0x58331002: /* ATI RS300_200 AGP bridge */
				IOLog("AGPATI: Chipset is RS300.\n");
				apsize_reg = ATI_RS300_APSIZE;
                 agpmode_reg = ATI_RS300_IG_AGPMODE;
                 break;
		 case 0x59501002: // ATI RS400 AGP bridge 
				IOLog("AGPATI: Chipset is RS400.\n");
				apsize_reg = ATI_RS300_APSIZE;
                 agpmode_reg = ATI_RS300_IG_AGPMODE;
                 break;

		 case 0x79301002: // ATI RS400 AGP bridge 
				IOLog("AGPATI: Chipset is RS480.\n");
				apsize_reg = ATI_RS300_APSIZE;
				agpmode_reg = ATI_RS300_IG_AGPMODE;
				break;
					  
         default:
                 IOLog("AGPATI: Chipset type not found. Propose RV100 variant\n");
				 apsize_reg = ATI_RS100_APSIZE;
                 agpmode_reg = ATI_RS100_IG_AGPMODE;
         };
	vendMast = master->configRead32(0) & 0xffff;
	switch(vendMast)
	{
		case 0x1002:
			IOLog("AGPATI: VideoCard by ATI\n");
			break;
		case 0x10de:
			IOLog("AGPATI: VideoCard by nVidia\n");
			break;
		default:
			IOLog("AGPATI: unknown VideoCard\n");
	}
			
			
	addrMask = 1 | ITA_COH; //- Slice AGP is always non-coherent
	isAMD = true; //false;  yes, ATI chipset works as AMD
	bridge->setMemoryEnable(true);
	IOPCIDevice * masterDevice = (IOPCIDevice *)master;
	RADio = 0;
	if(vendMast == 0x1002)  // don't tune nVidia
	{
		masterDevice->setMemoryEnable(true);
		//masterDevice->setIOEnable(true);
		for (UInt32 i = 0; (RADio = masterDevice->mapDeviceMemoryWithIndex(i)); i++)
		{
			RADio_base_phys = RADio->getPhysicalAddress();
			RADio_base = (volatile UInt8 *)RADio->getVirtualAddress();
			// Make sure we  select MMIO registers
			if (((RADio->getLength()) == 0x00010000) && (RADio_base_phys != 0))
					break;
		}
	}
	IOMemoryDescriptor *bridgeMemory = IOMemoryDescriptor::withPhysicalAddress(pciread32(hostSpace, ATI_GART_MMADDR), 0x100, kIODirectionInOut);
	if(bridgeMemory)
	{
	  mmio = bridgeMemory->map();
	  if(!mmio)
	  {
		IOLog("AGPATI: Memory not mapped\n");
		} else 
		{
			mmio_base_phys = mmio->getPhysicalAddress();
			mmio_base = (volatile UInt8 *)mmio->getVirtualAddress();
		}
	} else 
		IOLog("AGPATI: bridgeMemory not created\n");
					
// For trace new space
//	IOLog("AGPATI: map mmio %08lx to @%08lx\n", (UInt32)mmio_base_phys, (UInt32)mmio_base);

	apsizeValue = pciread16(hostSpace, targetAGPRegisters + APSIZE);
UInt32 curLength = (((apsizeValue & 0x3f) ^ 0x3f) + 1) << 22; //Slice VRAM independent
UInt32 MEM_CONF = 0x04000000;
UInt32 VRAMaddr = pciread32(hostSpace,0x90);
rVRAM = curLength / 2;
UInt32 tmp = pciread16(hostSpace, apsize_reg);
systemLength = 1 << (((tmp & 0xe) >> 1) + 25);
if(RADio) //Slice  - memory sizes
   {
   VRAM = INRAD(0x108);
   MEM_CONF = INRAD(0xf8);
   //systemLength = VRAM;
	UInt32 tom = INRAD(0x15c);
	VRAMaddr = tom << 16;
	rVRAM = ((tom | 0xffff) +1) - (tom << 16);
	//systemLength = rVRAM << 1;
	}

IOLog("AGPATI: aperLength = %08lx asked=%08lx (108)=%08lx (f8)=%08lx\n", systemLength, *length, VRAM, MEM_CONF);	

// Radeon Driver works with aperture 32Mb, 64Mb or 128Mb and no more
	if (systemLength > *length)
	{
		systemLength = *length;
		IOLog("AGP: Aperture Reduced To %ldMB\n", systemLength >> 20);
	}
	if (systemLength > 0x08000000)
	{
		systemLength = 0x08000000;
		IOLog("AGP: Aperture Reduced To %ldMB\n", systemLength >> 20);
	}
	
	if (systemLength != curLength)
	{
		values = A_SIZE_16(ati_generic_sizes);
		
		for (int i = 0; i < 7; i++) 
		{
			if ((systemLength >> 20) == (UInt16)values[i].size)
				apsizeValue = values[i].size_value;
		}		
		pciwrite16(hostSpace, targetAGPRegisters + APSIZE, apsizeValue);
	}
	tmp = 7;
	switch (systemLength){
		case 0x02000000:
			tmp = 1;
			break;
		case 0x04000000:
			tmp = 3;
			break;
		case 0x08000000:
			tmp = 5;
			break;
	}
	pciwrite32(hostSpace, apsize_reg, tmp);
	
	//New APBASE
	
	UInt32 currentApbase, newApbase;
	tmp = pciread32( hostSpace, APBASE); // + (rVRAM << 1);
	currentApbase = (tmp & 0xfffffff0); // + systemLength;
//	IOLog("AGPATI APBASE=%08lx\n", currentApbase);
	if(*address == 0)
		newApbase =  GetBestAPBASE(master, currentApbase);//currentApbase;
	else newApbase = *address;
	
	//newApbase = (VRAMaddr - 1) & 0xf0000000; // - (systemLength << 1);
	//newApbase = VRAMaddr - systemLength;  //v249
	//newApbase = VRAMaddr + MEM_CONF; //as in new 8500
	//newApbase = pciread32( hostSpace, 0x90) & 0xffff0000;
	if(newApbase != currentApbase)
	{
		// Slice - before write we need to enable it
		pciwrite16(hostSpace, targetAGPRegisters + APSIZE, 0xffff);
		pciwrite32(hostSpace, APBASE, 0);
		pciwrite16(hostSpace, targetAGPRegisters + APSIZE, apsizeValue);
		pciwrite32(hostSpace, APBASE, newApbase | (tmp & 0x0000000f));
	}
	
	MEM_CONF = ((newApbase + systemLength - 1) & 0xffff0000) + (newApbase >> 16);
	if(RADio)
	{
		OUTRAD(0x14c, MEM_CONF);
		//OUTRAD(0x130, INRAD(0x130) | (1<<23));  //bad rezult
		//OUTRAD(0x134, (VRAMaddr >> 16) | 0xffff0000);  //not writable
		//pciwrite32(hostSpace, 0xec, MEM_CONF);
		//IOLog("AGPATI: MC_AGP in ATI=%08lx\n",(UInt32)INRAD(0x14c));	
	}

	
	systemBase			=	newApbase;		
	initial_aperture	= systemLength;
//	entries				= (initial_aperture >> AGP_PAGE_SHIFT);
	IOLog("AGPATI: aperture [%08lx, %08lx]\n", systemBase, systemLength);
	if(systemLength < 0x02000000 || systemLength > 0x08000000)	  //Slice && -> ||
		IOLog("AGP: WARNING! Please set your aperture to 32, 64 or 128MB in your BIOS for more compatibility\n");

		*address = systemBase;
		*length = systemLength;


	return err;
}
//----------------------------------------------------------------------------

IOReturn 
AGPATI::InstallGATT(IOPhysicalAddress gattPhys)
{
	UInt32 temp;
	IOReturn err;
	//APSIZE  -- linux writing; Slice - nonsense
   temp = pciread32(hostSpace, apsize_reg);
	pciwrite32(hostSpace, apsize_reg, temp | 1);  //Gart_enable -- already done!
	//Flush0TLB();
	pciwrite32(hostSpace, agpmode_reg, 0x20000);  //ATI specific Linux shit non-documented
	/* documented bits are
		1<<18  -  half queue  size
		1<<19  -  probe outside GART
		1<<20  -  half buffer size
		*/

	if(mmio)
	{
		OUTREG(ATI_GART_FEATURE_ID, 0x60000 | (1<<10)); //Slice - yes!  Linux=0x60000
		// (1<<10) - AGP capability enable
		temp = INREG(ATI_GART_FEATURE_ID);
	//	OUTREG(ATI_GART_CACHE_ENTRY_CNTRL, 0xf);
	} 
		/*  AGPCTRL
			We have disabled the AGPCTRL.gtlben bit to prevent the translation before
			the configuration is complete. */
		pciwrite32(hostSpace, targetAGPRegisters + AGP_CTRL, 0x0000);
		temp = pciread16(hostSpace, 4);      // XXX: Magic reg# 
		pciwrite16(hostSpace, 4, temp | (1 << 14));  //Slice - nonsense
		/*  ATTBASE
			Now we write our GATT register with the physical address our our GART table.  */
		/*	        writel(agp_bridge->gatt_bus_addr, ati_generic_private.registers+ATI_GART_BASE);
					readl(ati_generic_private.registers+ATI_GART_BASE);     // PCI Posting. 
	*/

		UInt32 attTemp = (gattPhys << 12) | ((gattPhys & 0x0ff00000) >> 16);
		if(mmio)
		{
			OUTREG(ATI_GART_BASE, attTemp);
			//OUTREG(ATI_GART_CACHE_CNTRL, 0x1);
			//OUTREG(ATI_GART_CACHE_ENTRY_CNTRL, (3<<7) | 1);
//			IOLog("AGPATI: gartPhys written into host %08lx\n ",(UInt32)INREG(ATI_GART_BASE)); 
		} 
		/*  AGPCTRL
			We re-enable the gtlben bit to enable translation, even though it will be 
			deactivated in the commitAGPMemory method.*/
			UInt32	agpCtrl = pciread16(hostSpace, targetAGPRegisters + AGP_CTRL);
			pciwrite16(hostSpace, targetAGPRegisters + AGP_CTRL, agpCtrl | (3 << 7));
			
		 if(RADio)
		 {
			temp = INRAD(0x174); // & ~0x3f;
			//Slice -- this is AGP_CNTL. What to do with it?
			OUTRAD(0x174, temp | 0xe0000); //temp2 | (apsizeValue & 0x3f));  -- xf86
			//OUTRAD(0x174, temp | 0x20); // -- linux 2.6.17 Slice - non-documented bit
			//OUTRAD(0x174, temp | 0x800); // force full SBA
			//OUTRAD(0x174, 0);  //Linux tracing
			//UInt32 tom = INRAD(0x15c);
			//UInt32 VRAMaddr = tom << 16;
			//UInt32 rVRAM = ((tom | 0xffff) +1) - (tom << 16);

			//temp = ((VRAMaddr + rVRAM - 1) & ~0xffff) + (systemBase >> 16);
			//OUTRAD(0x40, INRAD(0x40) | (1<<24));
			//temp = 0xffff0000 + (systemBase >> 16);
			//OUTRAD(0x14c, temp);
			//OUTRAD(0x17c, 0x00400000);
			OUTRAD(0x910, 2); //IGP ??? non-documanted

			//OUTRAD(0xf8, systemLength);
			//OUTRAD(0xfc, 0x20000);
		 }
	//	Flush1TLB();
		
	// Trace after config
//	IOLog("AGPATI: trace after config\n");
			if(mmio)
		for(int i=0; i<0x1f; i+=16)
		{
			for(int j=0; j<15; j+=4)
				IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)INREG(i+j));  //for trace
			IOLog("\n");
		} 
 /*			
		if(RADio)
		for(int i=0x100; i<0x1ff; i+=16)
		{
			for(int j=0; j<15; j+=4)
				IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)INRAD(i+j));
			IOLog("\n");
		} 
*/
		IOLog("AGPATI trace PCI space Host dev=0\n");
		for(int i=0; i<0x2f; i+=16)
			{
				for(int j=0; j<15; j+=4)
					IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)pciread32(hostSpace,i+j));  //for trace
				IOLog("\n");
			} 
		for(int i=0xa0; i<0xff; i+=16)
		{
			for(int j=0; j<15; j+=4)
				IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)pciread32(hostSpace,i+j));  //for trace
			IOLog("\n");
		} 

		readyForPM = true;
		err = kIOReturnSuccess;
	
	return ( err );
}

//----------------------------------------------------------------------------

void AGPATI::Disable( IOAGPDevice * master )
{
	super::Disable(master);
}

//----------------------------------------------------------------------------
/*AMD document
GART_CACHE_CNTRL - RW - 32 bits - NBMCIND:0x2E 
Field Name Bits Default Description 
GART_CACHE_INVALIDATE_W (W) 0 0x0 When set to 1, the entire GART Directory and Table Cache 
is invalidated. When the operation is completed, set this bit to 0. 
0=No change 
1=Clear Tlb 
GART_CACHE_INVALIDATE_R (R) 0 0x0 When GART_CACHE_INVALIDATE_W is set to 1, 
GART_CACHE_INVALIDATE_R will toggle from 1 to 0 
when the operation is completed. 
0=Done Clear 
1=Clear Pending 
GART Cache Control Register 

 */
void AGPATI::Flush0TLB()
{
	//Don't touch!
//	if(mmio) 
//	{
//		OUTREG(ATI_GART_BASE, gartPhys);
//		temp = INREG(ATI_GART_CACHE_ENTRY_CNTRL);
//		OUTREG(ATI_GART_CACHE_CNTRL, 1); 
 //   readl(ati_generic_private.registers+ATI_GART_CACHE_CNTRL);      /* PCI Posting. */
//		(void)INREG(ATI_GART_CACHE_CNTRL); //ENTRY_
//	}
	//	temp = pciread32(hostSpace, targetAGPRegisters + AGP_CTRL);
//	pciwrite32(hostSpace, targetAGPRegisters + AGP_CTRL, 0);		

}
//----------------------------------------------------------------------------

void AGPATI::Flush1TLB()
{
//	UInt32 temp;
	if(mmio) 
	{
//		OUTREG(ATI_GART_BASE, gartPhys);
//		temp = INREG(ATI_GART_CACHE_ENTRY_CNTRL);
		OUTREG(ATI_GART_CACHE_CNTRL, 1); //ENTRY_
/*		for (UInt32 i = 0; i < flusht; i++)
		{
//			IODelay(10);
			temp = INREG(ATI_GART_CACHE_CNTRL) & 1;
			if(temp == 0) break;
		}
		OUTREG(ATI_GART_CACHE_CNTRL, 0);
*/
		
		UInt32 w;
		/* Make sure the hardware actually executed the flush. */
		do { 
			w = INREG(ATI_GART_CACHE_CNTRL);
		} while (w & 1);
		OUTREG(ATI_GART_CACHE_CNTRL, 0);
	}	
//	temp = pciread32(hostSpace, targetAGPRegisters + AGP_CTRL);
//	pciwrite32(hostSpace, targetAGPRegisters + AGP_CTRL, temp | (1 << 7) | (1 << 8));		


}

//----------------------------------------------------------------------------

void AGPATI::CheckTLB()
{
	if(mmio) 
	{
//		OUTREG(ATI_GART_BASE, gartPhys);
//		temp = INREG(ATI_GART_CACHE_ENTRY_CNTRL);
//		OUTREG(ATI_GART_CACHE_CNTRL, 1); //ENTRY_
 //   readl(ati_generic_private.registers+ATI_GART_CACHE_CNTRL);      /* PCI Posting. */
		IOLog("AGPATI: gartBase=%08lx\n",(UInt32)INREG(ATI_GART_BASE));
	}

}

//----------------------------------------------------------------------------
//static int IOLogCount = 0;
//static int cnt = 0;
//static UInt32 physOld = 0;
/*

IOReturn AGPATI::BindMemory(IOMemoryDescriptor * mem, IOByteCount agpOffset)
{	
//	    IOReturn		err = kIOReturnSuccess;
  //  UInt32		offset = 0;
	//UInt32	mem_phys;
	IOByteCount	len; //, len2;
	IOPhysicalAddress	phys, phys2;
//	IOByteCount         agpFlushStart, flushLength;
	//cnt++;
	if(0) 
	{
		OUTREG(ATI_GART_CACHE_CNTRL, 0); 
	}
	phys = mem->getPhysicalAddress();
	//mem_phys = phys;
	if(phys != physOld)
	{
	 IOLogCount = 0;
	 physOld = phys;
	// if(bridgeDev->BindMemory(memory,agpOffset) != 0)
	//		err = kIOReturnError;
	 } 
	
//	UInt32 temp;
//	temp = pciread32(hostSpace, targetAGPRegisters + AGP_CTRL);
//	pciwrite32(hostSpace, targetAGPRegisters + AGP_CTRL, temp & ~(1 << 7));		
	len = mem->getLength();
	UInt32 gartOffset = agpOffset >> 10;
//	agpFlushStart = gartOffset;
	//agpOffset >> 10;	//	/= (page_size / 4);
		
//	while( (phys = mem->getPhysicalSegment(offset, &len)))
//	{
		phys += agpOffset;
	//	offset += len;
	//	len = (len + PAGE_MASK) & ~PAGE_MASK;
		while( len > 0) 
		{
			OSWriteLittleInt32(gartArray, gartOffset, phys | addrMask);
			gartOffset += 4;
			phys += page_size;
			len -= page_size;
		}
//	}
//	flushLength = (gartOffset - agpFlushStart);
//	flush_dcache( ((vm_offset_t)gartArray)+agpFlushStart, flushLength, false);
	phys2 = OSReadLittleInt32(gartArray, gartOffset-4);
	
//	mem->performOperation(kIOMemoryIncoherentIOFlush, 0, len2); //mem->getLength() );
//	agp_flush_cache();
//	FlushTLB();
//	temp = pciread32(hostSpace, targetAGPRegisters + AGP_CTRL);
//	pciwrite32(hostSpace, targetAGPRegisters + AGP_CTRL, temp | (1 << 7) | (1 << 8));		

//	return( 0 );


//	UInt32 * RadVars;
//	phys2 = 0x12345678;

	UInt32 Regs[] = {0xf8, 0x100, 0x104,0x108, 0x148, 0x14c, 0x15c, 0x170, 0xf70,
			0x1d8, 0x1dc, 0x1e0, 0x1e4, 0x23c, 0x33c, 0x43c, 0x700, 0x774, 0x19e4, 0};
//	IOLog("AgpAti: Start with %s\n", provider->getName());
		if(RADio)
		{
			IOLog("AgpAti: trace after Radeon.kext\n");
			int i = 0;
			while(Regs[i])
			{
			  IOLog("(%08lx)=%08lx  ", Regs[i], (UInt32)INRAD(Regs[i]));
			  i++;
			  if(!(i & 3)) IOLog("\n");
			  }
		}
	return( 0 );

	mem_phys = mem->getPhysicalAddress();
	len2 = mem->getLength();
	//	assert( agpOffset < systemLength );
	agpOffset /= (page_size / 4);
		
	while( (phys = mem->getPhysicalSegment(offset, &len)))
	{
		offset += len;
		len = (len + PAGE_MASK) & ~PAGE_MASK;
		while( len > 0) 
		{
			phys |= 0x00000001;  //Slice - COH absent
			//phys2 = phys;
			OSWriteLittleInt32( gartArray, agpOffset, phys);
			phys2 = OSReadLittleInt32( gartArray, agpOffset);
			agpOffset += 4;
			phys += page_size;
			len -= page_size;
		}
	}

    len = OSReadLittleInt32( gartArray, agpOffset - 4 );  */
/*	if (IOLogCount < 4)
	{
		IOLog("AGPATI Bind memory %08lx length %08lx offset %08lx -> phys %08lx\n", (UInt32)mem_phys, (UInt32)len2, (UInt32)agpOffset, (UInt32)phys2);
		IOLogCount++;
	}
	
	if (IOLogCount == 2)
	{
		IOLog("AGPATI: trace at bind\n");
		if(RADio)
			for(int i=0; i<0x1ff; i+=16)
			{
				for(int j=0; j<15; j+=4)
					IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)INRAD(i+j));  //for trace
				IOLog("\n");
			} 
	}
	return kIOReturnSuccess;
}*/

//----------------------------------------------------------------------------
