/*
 * Copyright (c) 1998-2000 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#include <IOKit/IOLib.h>
#include <IOKit/platform/ApplePlatformExpert.h>
#include <IOKit/pci/IOPCIBridge.h>
#include <IOKit/IOMapper.h>

#include "IOAGPBridge.h"


#define INTEL_APSIZE    0xb4
#define APBASE	0x10
#define  PCI_BASE_ADDRESS_MEM_MASK      (~0x0fUL)
#define INTEL_ATTBASE_MASK 0x00000FFF
#define INTEL_I865_AGPM         0x51
#define INTEL_I865_ERRCMD		0xCA
#define INTEL_I865_ERRSTS       0xc8

#define CAPABILITY_PTR	0x34

// Then shift in AGP space
#define AGP_STAT			0x04
#define AGP_CMD				0x08
#define AGP_CTRL			0x10
#define APSIZE				0x14 
#define ATTBASE				0x18



/********************************************************************/
/*						INTEL DRIVER FUNCTIONS						*/
/*						 by joblo10 and Tom H						*/
/*						    August 8, 2006							*/	
/********************************************************************/

#define super IOAGPBridge

//----------------------------------------------------------------------------

AGPINTEL::AGPINTEL(IOService *provider)
{
	//Accessor for configRead and configWrite 
	bridge = (IOPCIDevice *)provider;
	
	//PCI Space of the host bridge
	hostSpace = bridge->space;
	AGPBSpace = hostSpace;
	hostSpace.s.deviceNum = 0;	
}

//----------------------------------------------------------------------------

const char* AGPINTEL::Match(UInt16 dev)
{
	//Slice - this is AGPBridge dev=1
	 switch (dev) {
		 case 0x1a31: //
				 return ("Intel 82845 host to AGP bridge");
		 case 0x3341: //
				 return ("Intel 82855 host to AGP bridge");
		 case 0x2532: //
				 return ("Intel 82860 host to AGP bridge");
		 case 0x2571: //
				 return ("Intel 82865 host to AGP bridge");
		 case 0x255d:
				 return ("Intel E7205 host to AGP bridge");
		 case 0x2579: //
				 return ("Intel 82875P host to AGP bridge");
		 case 0x2561:
				 return ("Intel 82845G host to AGP bridge");
		 case 0x3581:
				 return ("Intel 82855GM host to AGP bridge");
	 };

	 return NULL;
} 

//----------------------------------------------------------------------------

static struct 
aper_size_info_16 intel_generic_sizes[] =
{
         {256, 65536, 6, 0},
         {128, 32768, 5, 0x20},
         {64, 16384, 4, 0x30},
         {32, 8192, 3, 0x38},
         {16, 4096, 2, 0x3c},
         {8, 2048, 1, 0x3e},
         {4, 1024, 0, 0x3f}
};

UInt32 AGPINTEL::GetBestAPBASE(IOAGPDevice *_master, UInt32 currentAperBase)
{
	IODeviceMemory *mem = getVRAMMemory(_master);
	UInt32 VRAMBase, VRAMLength, gap, temp;
	
	VRAMBase = mem->getPhysicalAddress();
	VRAMLength = mem->getLength()>>1;
	//temp = currentAperBase;
	temp = VRAMBase + VRAMLength;
	//First, verify if there's a GAP between VRAM and AGP aperture
	gap = currentAperBase - temp; //(VRAMBase + VRAMLength);
	if(currentAperBase > temp) //(VRAMBase + VRAMLength))
	{
		//gap exist and agp aperture is higher than VRAM, APBASE will be reduced
		temp = currentAperBase - gap;
		IOLog("AGP: Gap exists, aperBase corrected\n");
		//verify it's a valid value
		if(temp & 0x00FFFFFF)
		{
			IOLog("AGP: New APBASE invalid, return default value\n");
			return currentAperBase;
		}
	}
		
	return temp;
	
}

//----------------------------------------------------------------------------

IOReturn AGPINTEL::Configure(IOAGPDevice *master, IOPhysicalAddress * address, IOPhysicalLength * length)
{
 //   UInt32		agpCtrl;
	aper_size_info_16 *values;
	IOReturn err;
	
	addrMask = 1 | ITA_COH;
		bridge->setMemoryEnable(true);
		isAMD = false;
	apsizeValue = pciread8(hostSpace, AGP_INTEL_APSIZE);
	systemLength	= (((apsizeValue & 0x3f) ^ 0x3f) + 1) << 22;
	
	// On system with high aperture, the IOGraphics makes strange problems.
	// Aperture requested by OSX is always 128MB, so we use that default setting here,
	// to make everyone happy. There were reports that this hack make ATIRadeon8500 to 
	// load.
	// Slice: ATIRadeon8500 work only with 32, 64 or 128Mb aperture
	if (systemLength > 0x08000000)
	{
		systemLength = 0x08000000;
		IOLog("AGP: Aperture Reduced To %ldMB\n", systemLength >> 20);
	}

	if (systemLength > *length)
	{
		systemLength = *length;
		IOLog("AGP: Aperture Reduced To %ldMB\n", systemLength >> 20);
		values = A_SIZE_16(intel_generic_sizes);
		
		for (int i = 0; i < 7; i++) 
		{
			if ((systemLength >> 20) == (UInt16)values[i].size)
				apsizeValue = values[i].size_value;
		}		
		pciwrite8(hostSpace, AGP_INTEL_APSIZE, apsizeValue);
	}
	
	//New APBASE
	UInt32 currentApbase, newApbase, tmp;
	tmp =  pciread32( hostSpace, APBASE );
	currentApbase = tmp & 0xfffffff0;
	// 10.03.2008
	if(*address == 0)
		newApbase = GetBestAPBASE(master, currentApbase);
	else newApbase = *address;
	
	if(newApbase != currentApbase)
	{
		// Slice - before write we need to enable it
		pciwrite8(hostSpace, AGP_INTEL_APSIZE, 0xff);
		pciwrite32(hostSpace, APBASE, 0);
		pciwrite8(hostSpace, AGP_INTEL_APSIZE, apsizeValue);
		pciwrite32(hostSpace, APBASE, newApbase | (tmp & 0x0000000f));
	}

	
	systemBase			= newApbase;
	initial_aperture	= systemLength;
	//entries				= initial_aperture >> AGP_PAGE_SHIFT;
	
	if( !systemLength)
	{
		IOLog("AGP: ERROR! Could not find aperture size from registers\n");
		return -1;
	}
	else if(systemLength < 0x02000000 || systemLength > 0x08000000)	 //Slice && -> ||
		IOLog("AGP: WARNING! Please set your aperture to 64 or 128MB in your BIOS for more compatibility\n");
		*address = systemBase;
		*length = systemLength;
		IOLog("AGPINTEL: aperture [%08lx, %08lx]\n", systemBase, systemLength);

		err = kIOReturnSuccess;

	return ( err );
}
//----------------------------------------------------------------------------

IOReturn 
AGPINTEL::InstallGATT(IOPhysicalAddress gattPhys)
{
	UInt32 type = pciread32(hostSpace, 0);
	UInt32 tmp;
	IOReturn err;
		/*  AGPCTRL
			We have disabled the AGPCTRL.gtlben bit to prevent the translation before
			the configuration is complete. */
		//pciwrite32(hostSpace, AGP_INTEL_AGPCTRL, 0x0000);
		Flush0TLB();

         /* Install the gatt. */
         pciwrite32(hostSpace, AGP_INTEL_ATTBASE, gattPhys << 12);
		 tmp = pciread32(hostSpace, AGP_INTEL_ATTBASE); //Slice - pci posting
 
         /* Enable the GLTB and setup the control register. */
         switch (type) {
         case 0x71908086: /* 440LX/EX */
                 pciwrite32(hostSpace, AGP_INTEL_AGPCTRL, 0x2080);
                 break;
         case 0x71808086: /* 440BX */
                 /*
                  * XXX: Should be 0xa080?  Bit 9 is undefined, and
                  * bit 13 being on and bit 15 being clear is illegal.
                  */
                 pciwrite32(hostSpace, AGP_INTEL_AGPCTRL, 0x2280);
                 break;
         default:
                 tmp = pciread32(hostSpace, AGP_INTEL_AGPCTRL);
                 pciwrite32(hostSpace, AGP_INTEL_AGPCTRL, tmp | 0x80);
         }
 
         /* Enable things, clear errors etc. */
  /*       switch (type) {
         case 0x1a218086: // i840 
         case 0x25308086: // i850 
         case 0x25318086: // i860 
                 pciwrite16(hostSpace, AGP_INTEL_MCHCFG,
                                  (pciread16(hostSpace, AGP_INTEL_MCHCFG)
                                   | (1 << 9)));
                 break;
 
         case 0x25008086: // i820 
         case 0x25018086: // i820 
                 pciwrite8(hostSpace, AGP_INTEL_I820_RDCR,
                                  (pciread8(hostSpace, AGP_INTEL_I820_RDCR)
                                   | (1 << 1)));
                 break;
 
         case 0x1a308086: // i845 
         case 0x33408086: // i855 
		 case 0x33418086: // i855 
         case 0x35808086: // i855GM 
         case 0x255d8086: // E7205 
         case 0x25508086: // E7505 
         case 0x25708086: // i865 
         case 0x25788086: // i875P 
         case 0x25608086: // i845G 
                 pciwrite8(hostSpace, AGP_INTEL_I845_MCHCFG,
                                  (pciread8(hostSpace, AGP_INTEL_I845_MCHCFG)
                                   | (1 << 1)));
                 break;
 
         default: // Intel Generic (maybe) 
                 pciwrite32(hostSpace, AGP_INTEL_NBXCFG,
                                  (pciread32(hostSpace, AGP_INTEL_NBXCFG)
                                   & ~(1 << 10)) | (1 << 9));
         }
 */
         switch (type) {
         case 0x1a218086: /* i840 */
                 pciwrite16(hostSpace, AGP_INTEL_I8XX_ERRSTS, 0xc000);
                 break;
 
         case 0x25008086: /* i820 */
         case 0x25018086: /* i820 */
         case 0x1a308086: /* i845 */
		 case 0x25308086: /* i850 */
         case 0x33408086: /* i855 */
		 case 0x35808086: /* i855G */
         case 0x255d8086: /* E7205 */
         case 0x25508086: /* E7505 */
         case 0x25318086: /* i860 */
         case 0x25708086: /* i865 */
         case 0x25788086: /* i875P */
         case 0x25608086: /* i845G */
                 pciwrite16(hostSpace, AGP_INTEL_I8XX_ERRSTS, 0x00ff);
                 break;
 
         default: /* Intel Generic (maybe) */
                 pciwrite8(hostSpace, AGP_INTEL_ERRSTS + 1, 7);
		}
				
		
		
			/*  AGPM 
			Enable access to the aperture, when bit = 0, aperture is not 
			accessible by the chipset, target and master devices.*/	
		UInt8 temp2 = pciread8(hostSpace, kiAGPM);
		pciwrite8(hostSpace, kiAGPM, temp2 | (1 << 1));
		
		Flush1TLB();
	

		/*IOLog("AGP: Registers after configuration:\n\t "
									"AGPCTRL: 0X%08lx, ATTBASE: 0X%08lx, "
									"AGPM: 0X%02lx, ERRCMD: 0X%04lx, ERRSTS: 0X%04lx\n", pciread32( target, kiAGPCTRL ), 
									pciread32( target, kiATTBASE ), pciread8(target, INTEL_I865_AGPM),
									pciread16(target, INTEL_I865_ERRCMD), errsts );*/
		//Slice Trace PCI space
		IOLog("AGPINTEL trace PCI space\n");
		for(int i=0; i<0x2f; i+=16)
			{
				for(int j=0; j<15; j+=4)
					IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)pciread32(hostSpace,i+j));  //for trace
				IOLog("\n");
			} 


		readyForPM = true;
		err = kIOReturnSuccess;

	return ( err );
}

//----------------------------------------------------------------------------

void AGPINTEL::Disable( IOAGPDevice * master )
{
	// disable aperture access
	UInt32 devID = pciread32(hostSpace, 0);
	switch ( devID ) {
	case 0x25008086: /* i820 */
	case 0x25018086: /* i820 */
	case 0x1a308086: /* i845 */
	case 0x25608086: /* i845G */
	case 0x33408086: /* i855 */
	case 0x35808086: /* i855GM */
	case 0x25708086: /* i865 */
	case 0x25788086: /* i875P */
		pciwrite8(hostSpace, kiAGPM,
			pciread8(hostSpace, kiAGPM ) & ~(1 << 1) );
		break;
		
	case 0x1a218086: /* i840 */
	case 0x25308086: /* i850 */
	case 0x25318086: /* i860 */
	case 0x255d8086: /* E7205 */
	case 0x25508086: /* E7505 */
		pciwrite16(hostSpace, AGP_INTEL_MCHCFG,
			pciread16(hostSpace, AGP_INTEL_MCHCFG ) & ~(1 << 9) );
		break;
		
	default: /* Intel Generic (maybe) */
		pciwrite32(hostSpace, AGP_INTEL_NBXCFG,
			pciread32(hostSpace, AGP_INTEL_NBXCFG ) & ~(1 << 9) );
	}
	
	//pciwrite32(hostSpace, kIntelAttBase, 0 );

	super::Disable(master);
}

//----------------------------------------------------------------------------

void AGPINTEL::Flush0TLB()
{
	UInt32 temp;
	
	temp = pciread32(hostSpace, AGP_INTEL_AGPCTRL);
	pciwrite32(hostSpace, AGP_INTEL_AGPCTRL, temp & ~(1 << 7));
}
//----------------------------------------------------------------------------

void AGPINTEL::Flush1TLB()
{
	UInt32 temp;
	temp = pciread32(hostSpace, AGP_INTEL_AGPCTRL);
	pciwrite32(hostSpace, AGP_INTEL_AGPCTRL, temp | (3 << 7));
}

//----------------------------------------------------------------------------

void AGPINTEL::CheckTLB()
{
}
//----------------------------------------------------------------------------
//static int IOLogCount = 0;
/*

IOReturn AGPINTEL::BindMemory(IOMemoryDescriptor * mem, IOByteCount agpOffset)
{	
 //   UInt32		offset = 0;
	//UInt32	mem_phys;
	IOByteCount	len; //, len2;
	IOPhysicalAddress	phys, phys2;
	IOByteCount         agpFlushStart, flushLength;
	phys = mem->getPhysicalAddress();
 	UInt32 temp;
	temp = pciread32(hostSpace, AGP_INTEL_AGPCTRL);
	pciwrite32(hostSpace, AGP_INTEL_AGPCTRL, temp & ~(1 << 7));

	len = mem->getLength();
	UInt32 gartOffset = agpOffset >> 10;
	agpFlushStart = gartOffset;
//	while( (phys = mem->getPhysicalSegment(offset, &len)))
//	{
		phys += agpOffset;
	//	offset += len;
	//	len = (len + PAGE_MASK) & ~PAGE_MASK;
		while( len > 0) 
		{
			OSWriteLittleInt32(gartArray, gartOffset, phys | addrMask);
			gartOffset += 4;
			phys += page_size;
			len -= page_size;
		}
//	}
	flushLength = (gartOffset - agpFlushStart);
	flush_dcache( ((vm_offset_t)gartArray)+agpFlushStart, flushLength, false);
	phys2 = OSReadLittleInt32(gartArray, gartOffset-4);
//	mem->performOperation(kIOMemoryIncoherentIOFlush, 0, len2); //mem->getLength() );
//	agp_flush_cache();
	FlushTLB();



	return( 0 );

}
*/
//----------------------------------------------------------------------------
